@@ -1,6 +1,8 @@
 /* frdp-session.c
  *
  * Copyright (C) 2018 Felipe Borges <felipeborges@gnome.org>
+ * Copyright (C) 2019 Armin Novak <akallabeth@posteo.net>
+ * Copyright (C) 2023 Marek Kasik <mkasik@redhat.com>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,12 +23,29 @@
 #include <freerdp/locale/keyboard.h>
 #include <freerdp/freerdp.h>
 #include <freerdp/gdi/gdi.h>
+#include <freerdp/gdi/video.h>
+#include <freerdp/gdi/gfx.h>
+#include <freerdp/client/channels.h>
+#include <freerdp/client/cmdline.h>
+#include <freerdp/client/channels.h>
+#include <freerdp/client/rdpei.h>
+#include <freerdp/client/tsmf.h>
+#include <freerdp/client/rail.h>
+#include <freerdp/client/cliprdr.h>
+#include <freerdp/client/rdpgfx.h>
+#include <freerdp/client/encomsp.h>
+#include <freerdp/client/disp.h>
+#include <freerdp/client/geometry.h>
+#include <freerdp/client/video.h>
+#include <freerdp/channels/disp.h>
 #include <gdk/gdk.h>
 #include <gio/gio.h>
 #include <gtk/gtk.h>
 #include <math.h>
 
 #include "frdp-session.h"
+#include "frdp-context.h"
+#include "frdp-channel-display-control.h"
 
 #define SELECT_TIMEOUT 50
 #define FRDP_CONNECTION_THREAD_MAX_ERRORS 10
@@ -46,8 +65,7 @@ struct _FrdpSessionPrivate
   cairo_surface_t *surface;
   cairo_format_t cairo_format;
   gboolean scaling;
-  double scale_x;
-  double scale_y;
+  double scale;
   double offset_x;
   double offset_y;
 
@@ -63,6 +81,10 @@ struct _FrdpSessionPrivate
   gboolean show_cursor;
   gboolean cursor_null;
   frdpPointer *cursor;
+
+  /* Channels */
+  FrdpChannelDisplayControl *display_control_channel;
+  gboolean                   monitor_layout_supported;
 };
 
 G_DEFINE_TYPE_WITH_PRIVATE (FrdpSession, frdp_session, G_TYPE_OBJECT)
@@ -75,7 +97,8 @@ enum
   PROP_USERNAME,
   PROP_PASSWORD,
   PROP_DISPLAY,
-  PROP_SCALING
+  PROP_SCALING,
+  PROP_MONITOR_LAYOUT_SUPPORTED
 };
 
 enum
@@ -84,18 +107,12 @@ enum
   RDP_CONNECTED,
   RDP_DISCONNECTED,
   RDP_AUTH_FAILURE,
+  RDP_CHANNEL_CONNECTED,
   LAST_SIGNAL
 };
 
 static guint signals[LAST_SIGNAL];
 
-struct frdp_context
-{
-  rdpContext context;
-  FrdpSession *self;
-};
-typedef struct frdp_context frdpContext;
-
 static void
 frdp_session_update_mouse_pointer (FrdpSession  *self)
 {
@@ -129,22 +146,22 @@ frdp_session_update_mouse_pointer (FrdpSession  *self)
     cursor = gdk_cursor_new_from_name (display, "default");
   else {
     rdpPointer *pointer = &priv->cursor->pointer;
-    double x = priv->cursor->pointer.xPos * priv->scale_x;
-    double y = priv->cursor->pointer.yPos * priv->scale_y;
-    double w = pointer->width * priv->scale_x;
-    double h = pointer->height * priv->scale_y;
+    double x = priv->cursor->pointer.xPos * priv->scale;
+    double y = priv->cursor->pointer.yPos * priv->scale;
+    double w = pointer->width * priv->scale;
+    double h = pointer->height * priv->scale;
     cairo_surface_t *surface;
     cairo_t *cairo;
 
     if (!self->priv->scaling) {
-      self->priv->scale_x = self->priv->scale_y = 1.0;
+      self->priv->scale = 1.0;
     }
 
     /* Scale the source image according to current settings. */
     surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, w, h);
     cairo = cairo_create (surface);
 
-    cairo_scale(cairo, self->priv->scale_x, self->priv->scale_y);
+    cairo_scale(cairo, self->priv->scale, self->priv->scale);
     cairo_set_source_surface (cairo, priv->cursor->data, 0, 0);
     cairo_paint (cairo);
 
@@ -173,9 +190,8 @@ static void
 create_cairo_surface (FrdpSession *self)
 {
   FrdpSessionPrivate *priv = self->priv;
-  rdpSettings *settings;
-  rdpGdi *gdi;
-  gint stride;
+  rdpGdi             *gdi;
+  gint                stride;
 
   if (priv->surface != NULL) {
     cairo_surface_mark_dirty (priv->surface);
@@ -184,12 +200,10 @@ create_cairo_surface (FrdpSession *self)
   }
 
   gdi = priv->freerdp_session->context->gdi;
-  settings = priv->freerdp_session->context->settings;
-  gdi_resize (gdi, settings->DesktopWidth, settings->DesktopHeight);
 
   gtk_widget_set_size_request (priv->display,
-                               settings->DesktopWidth,
-                               settings->DesktopHeight);
+                               gdi->width,
+                               gdi->height);
   stride = cairo_format_stride_for_width (priv->cairo_format, gdi->width);
   self->priv->surface =
       cairo_image_surface_create_for_data ((unsigned char*) gdi->primary_buffer,
@@ -200,6 +214,22 @@ create_cairo_surface (FrdpSession *self)
   cairo_surface_flush (priv->surface);
 }
 
+static gboolean
+frdp_desktop_resize (rdpContext *context)
+{
+  FrdpSession *self = ((frdpContext *) context)->self;
+  rdpGdi      *gdi = context->gdi;
+
+  if (gdi_resize (gdi,
+                  context->settings->DesktopWidth,
+                  context->settings->DesktopHeight)) {
+    create_cairo_surface (self);
+    return TRUE;
+  } else {
+    return FALSE;
+  }
+}
+
 static void
 frdp_session_configure_event (GtkWidget *widget,
                               GdkEvent  *event,
@@ -207,10 +237,11 @@ frdp_session_configure_event (GtkWidget *widget,
 {
   FrdpSession *self = (FrdpSession*) user_data;
   FrdpSessionPrivate *priv = self->priv;
-  GtkScrolledWindow *scrolled;
+  GtkWidget *scrolled;
   rdpSettings *settings;
   rdpGdi *gdi;
-  double width, height;
+  double width, height, widget_ratio, server_ratio;
+  gboolean allow_resize;
 
   if (priv->freerdp_session == NULL)
     return;
@@ -226,18 +257,33 @@ frdp_session_configure_event (GtkWidget *widget,
 
   if (priv->freerdp_session->settings == NULL)
     return;
+
   settings = priv->freerdp_session->settings;
 
-  if (priv->scaling) {
-    priv->scale_x = width / settings->DesktopWidth;
-    priv->scale_y = height / settings->DesktopHeight;
+  g_object_get (G_OBJECT (widget), "allow-resize", &allow_resize, NULL);
 
-    priv->offset_x = (width - settings->DesktopWidth * priv->scale_x) / 2.0;
-    priv->offset_y = (height - settings->DesktopHeight * priv->scale_y) / 2.0;
-
-    gtk_widget_set_size_request (priv->display, settings->DesktopWidth, settings->DesktopHeight);
+  if (allow_resize) {
+    if (settings->DesktopWidth != gtk_widget_get_allocated_width (scrolled) ||
+        settings->DesktopHeight != gtk_widget_get_allocated_height (scrolled)) {
+      frdp_channel_display_control_resize_display (priv->display_control_channel,
+                                                   width,
+                                                   height);
+    }
   } else {
-    gtk_widget_set_size_request (priv->display, gdi->width, gdi->height);
+    if (priv->scaling) {
+        widget_ratio = height > 0 ? width / height : 1.0;
+        server_ratio = settings->DesktopHeight > 0 ? (double) settings->DesktopWidth / settings->DesktopHeight : 1.0;
+
+        if (widget_ratio > server_ratio)
+          self->priv->scale = height / settings->DesktopHeight;
+        else
+          self->priv->scale = width / settings->DesktopWidth;
+
+        self->priv->offset_x = (width - settings->DesktopWidth * self->priv->scale) / 2.0;
+        self->priv->offset_y = (height - settings->DesktopHeight * self->priv->scale) / 2.0;
+    } else {
+      gtk_widget_set_size_request (priv->display, gdi->width, gdi->height);
+    }
   }
 }
 
@@ -259,17 +305,21 @@ frdp_session_draw (GtkWidget *widget,
   if (!self->priv->is_connected)
     return FALSE;
 
-  create_cairo_surface (self);
+  if (self->priv->surface == NULL ||
+      (self->priv->freerdp_session->context->gdi->width != cairo_image_surface_get_width (self->priv->surface) ||
+       self->priv->freerdp_session->context->gdi->height != cairo_image_surface_get_height (self->priv->surface))) {
+    create_cairo_surface (self);
+  }
 
   if (self->priv->scaling) {
       cairo_translate (cr, self->priv->offset_x, self->priv->offset_y);
-      cairo_scale (cr, self->priv->scale_x, self->priv->scale_y);
+      cairo_scale (cr, self->priv->scale, self->priv->scale);
   }
 
   cairo_set_source_surface (cr, self->priv->surface, 0, 0);
   cairo_paint (cr);
 
-  frdp_display_set_scaling (self->priv->display, self->priv->scaling);
+  frdp_display_set_scaling (FRDP_DISPLAY (self->priv->display), self->priv->scaling);
 
   return TRUE;
 }
@@ -314,15 +364,121 @@ frdp_authenticate (freerdp  *freerdp_session,
                                     domain);
 }
 
+static void
+caps_set (FrdpChannelDisplayControl *channel,
+          gpointer                   user_data)
+{
+  FrdpSession *session = user_data;
+
+  g_object_set (G_OBJECT (session), "monitor-layout-supported", TRUE, NULL);
+}
+
+static void
+frdp_on_channel_connected_event_handler (void                      *context,
+                                         ChannelConnectedEventArgs *e)
+{
+  frdpContext        *ctx = (frdpContext *) context;
+  FrdpSession        *session = ctx->self;
+  FrdpSessionPrivate *priv = frdp_session_get_instance_private (session);
+
+  if (strcmp (e->name, RDPEI_DVC_CHANNEL_NAME) == 0) {
+    // TODO Touch input redirection
+  } else if (strcmp (e->name, DISP_DVC_CHANNEL_NAME) == 0) {
+    g_clear_object (&priv->display_control_channel);
+
+    priv->display_control_channel = g_object_new (FRDP_TYPE_CHANNEL_DISPLAY_CONTROL,
+                                                  "session", session,
+                                                  "display-client-context", (DispClientContext *) e->pInterface,
+                                                  NULL);
+    g_signal_connect (priv->display_control_channel, "caps-set", G_CALLBACK (caps_set), session);
+  } else if (strcmp (e->name, TSMF_DVC_CHANNEL_NAME) == 0) {
+    // TODO Old windows 7 multimedia redirection
+  } else if (strcmp (e->name, RDPGFX_DVC_CHANNEL_NAME) == 0) {
+    gdi_graphics_pipeline_init (ctx->context.gdi, (RdpgfxClientContext *) e->pInterface);
+  } else if (strcmp (e->name, RAIL_SVC_CHANNEL_NAME) == 0) {
+    // TODO Remote application
+  } else if (strcmp (e->name, CLIPRDR_SVC_CHANNEL_NAME) == 0) {
+    // TODO Clipboard redirection channel
+  } else if (strcmp (e->name, ENCOMSP_SVC_CHANNEL_NAME) == 0) {
+    // TODO Multiparty channel
+  } else if (strcmp (e->name, GEOMETRY_DVC_CHANNEL_NAME) == 0) {
+    gdi_video_geometry_init (ctx->context.gdi, (GeometryClientContext *) e->pInterface);
+  } else if (strcmp (e->name, VIDEO_CONTROL_DVC_CHANNEL_NAME) == 0) {
+    gdi_video_control_init (ctx->context.gdi, (VideoClientContext *) e->pInterface);
+  } else if (strcmp (e->name, VIDEO_DATA_DVC_CHANNEL_NAME) == 0) {
+    gdi_video_data_init (ctx->context.gdi, (VideoClientContext *) e->pInterface);
+  }
+}
+
+static void
+frdp_on_channel_disconnected_event_handler (void                         *context,
+                                            ChannelDisconnectedEventArgs *e)
+{
+  frdpContext        *ctx = (frdpContext *) context;
+  FrdpSession        *session = ctx->self;
+  FrdpSessionPrivate *priv = frdp_session_get_instance_private (session);
+
+  if (strcmp (e->name, RDPEI_DVC_CHANNEL_NAME) == 0) {
+    // TODO Touch input redirection
+  } else if (strcmp (e->name, DISP_DVC_CHANNEL_NAME) == 0) {
+    g_clear_object (&priv->display_control_channel);
+  } else if (strcmp (e->name, TSMF_DVC_CHANNEL_NAME) == 0) {
+    // TODO Old windows 7 multimedia redirection
+  } else if (strcmp (e->name, RDPGFX_DVC_CHANNEL_NAME) == 0) {
+    gdi_graphics_pipeline_uninit (ctx->context.gdi, (RdpgfxClientContext *) e->pInterface);
+  } else if (strcmp (e->name, RAIL_SVC_CHANNEL_NAME) == 0) {
+    // TODO Remote application
+  } else if (strcmp (e->name, CLIPRDR_SVC_CHANNEL_NAME) == 0) {
+    // TODO Clipboard redirection channel
+  } else if (strcmp (e->name, ENCOMSP_SVC_CHANNEL_NAME) == 0) {
+    // TODO Multiparty channel
+  } else if (strcmp (e->name, GEOMETRY_DVC_CHANNEL_NAME) == 0) {
+    gdi_video_geometry_uninit (ctx->context.gdi, (GeometryClientContext *) e->pInterface);
+  } else if (strcmp (e->name, VIDEO_CONTROL_DVC_CHANNEL_NAME) == 0) {
+    gdi_video_control_uninit (ctx->context.gdi, (VideoClientContext *) e->pInterface);
+  } else if (strcmp (e->name, VIDEO_DATA_DVC_CHANNEL_NAME) == 0) {
+    gdi_video_data_uninit (ctx->context.gdi, (VideoClientContext *) e->pInterface);
+  }
+}
+
 static gboolean
 frdp_pre_connect (freerdp *freerdp_session)
 {
-  rdpSettings *settings;
+  rdpSettings *settings = freerdp_session->settings;
+  rdpContext *context = freerdp_session->context;
 
-  settings = freerdp_session->context->settings;
+  settings->OrderSupport[NEG_DSTBLT_INDEX] = TRUE;
+  settings->OrderSupport[NEG_PATBLT_INDEX] = TRUE;
+  settings->OrderSupport[NEG_SCRBLT_INDEX] = TRUE;
+  settings->OrderSupport[NEG_OPAQUE_RECT_INDEX] = TRUE;
+  settings->OrderSupport[NEG_DRAWNINEGRID_INDEX] = FALSE;
+  settings->OrderSupport[NEG_MULTIDSTBLT_INDEX] = FALSE;
+  settings->OrderSupport[NEG_MULTIPATBLT_INDEX] = FALSE;
+  settings->OrderSupport[NEG_MULTISCRBLT_INDEX] = FALSE;
+  settings->OrderSupport[NEG_MULTIOPAQUERECT_INDEX] = TRUE;
+  settings->OrderSupport[NEG_MULTI_DRAWNINEGRID_INDEX] = FALSE;
+  settings->OrderSupport[NEG_LINETO_INDEX] = TRUE;
+  settings->OrderSupport[NEG_POLYLINE_INDEX] = TRUE;
+  settings->OrderSupport[NEG_MEMBLT_INDEX] = TRUE;
+  settings->OrderSupport[NEG_MEM3BLT_INDEX] = FALSE;
+  settings->OrderSupport[NEG_MEMBLT_V2_INDEX] = TRUE;
+  settings->OrderSupport[NEG_MEM3BLT_V2_INDEX] = FALSE;
+  settings->OrderSupport[NEG_SAVEBITMAP_INDEX] = FALSE;
+  settings->OrderSupport[NEG_GLYPH_INDEX_INDEX] = TRUE;
+  settings->OrderSupport[NEG_FAST_INDEX_INDEX] = TRUE;
+  settings->OrderSupport[NEG_FAST_GLYPH_INDEX] = FALSE;
+  settings->OrderSupport[NEG_POLYGON_SC_INDEX] = FALSE;
+  settings->OrderSupport[NEG_POLYGON_CB_INDEX] = FALSE;
+  settings->OrderSupport[NEG_ELLIPSE_SC_INDEX] = FALSE;
+  settings->OrderSupport[NEG_ELLIPSE_CB_INDEX] = FALSE;
 
-  settings->RemoteFxCodec = TRUE;
-  settings->ColorDepth = 32;
+  PubSub_SubscribeChannelConnected (context->pubSub,
+                                    frdp_on_channel_connected_event_handler);
+  PubSub_SubscribeChannelDisconnected (context->pubSub,
+                                       frdp_on_channel_disconnected_event_handler);
+
+  if (!freerdp_client_load_addins (context->channels, settings))
+    return FALSE;
 
   return TRUE;
 }
@@ -358,13 +514,13 @@ frdp_end_paint (rdpContext *context)
   priv = self->priv;
 
   if (priv->scaling) {
-      pos_x = self->priv->offset_x + x * priv->scale_x;
-      pos_y = self->priv->offset_y + y * priv->scale_y;
+      pos_x = self->priv->offset_x + x * priv->scale;
+      pos_y = self->priv->offset_y + y * priv->scale;
       gtk_widget_queue_draw_area (priv->display,
                                   floor (pos_x),
                                   floor (pos_y),
-                                  ceil (pos_x + w * priv->scale_x) - floor (pos_x),
-                                  ceil (pos_y + h * priv->scale_y) - floor (pos_y));
+                                  ceil (pos_x + w * priv->scale) - floor (pos_x),
+                                  ceil (pos_y + h * priv->scale) - floor (pos_y));
   } else {
     gtk_widget_queue_draw_area (priv->display, x, y, w, h);
   }
@@ -375,9 +531,14 @@ frdp_end_paint (rdpContext *context)
 static gboolean
 frdp_post_connect (freerdp *freerdp_session)
 {
+  rdpSettings *settings;
+  rdpContext *context;
   FrdpSession *self = ((frdpContext *) freerdp_session->context)->self;
   guint32 color_format;
+  ResizeWindowEventArgs e;
 
+  context = freerdp_session->context;
+  settings = context->settings;
   switch (frdp_session_get_best_color_depth (self)) {
     case 32:
       color_format = PIXEL_FORMAT_BGRA32;
@@ -402,10 +563,34 @@ frdp_post_connect (freerdp *freerdp_session)
 
   freerdp_session->update->BeginPaint = frdp_begin_paint;
   freerdp_session->update->EndPaint = frdp_end_paint;
+  freerdp_session->update->DesktopResize = frdp_desktop_resize;
+
+  EventArgsInit(&e, "frdp");
+	e.width = settings->DesktopWidth;
+	e.height = settings->DesktopHeight;
+	PubSub_OnResizeWindow(context->pubSub, freerdp_session->context, &e);
+
+  create_cairo_surface (self);
 
   return TRUE;
 }
 
+static void
+frdp_post_disconnect (freerdp *instance)
+{
+  rdpContext *context;
+
+  if (!instance || !instance->context)
+    return;
+
+  context = instance->context;
+  PubSub_UnsubscribeChannelConnected (context->pubSub,
+                                      frdp_on_channel_connected_event_handler);
+  PubSub_UnsubscribeChannelDisconnected (context->pubSub,
+                                         frdp_on_channel_disconnected_event_handler);
+  gdi_free (instance);
+}
+
 static gboolean
 idle_close (gpointer user_data)
 {
@@ -474,12 +659,16 @@ static void
 frdp_session_init_freerdp (FrdpSession *self)
 {
   FrdpSessionPrivate *priv = self->priv;
-  rdpSettings *settings;
+  rdpSettings        *settings;
+  gchar              *collections[1];
+  gchar              *build_options;
+  int                 count = 1;
 
   /* Setup FreeRDP session */
   priv->freerdp_session = freerdp_new ();
   priv->freerdp_session->PreConnect = frdp_pre_connect;
   priv->freerdp_session->PostConnect = frdp_post_connect;
+  priv->freerdp_session->PostDisconnect = frdp_post_disconnect;
   priv->freerdp_session->Authenticate = frdp_authenticate;
   priv->freerdp_session->VerifyCertificate = frdp_certificate_verify;
   priv->freerdp_session->VerifyChangedCertificate = frdp_changed_certificate_verify;
@@ -512,9 +701,27 @@ frdp_session_init_freerdp (FrdpSession *self)
   settings->DesktopResize = TRUE;
   settings->DynamicResolutionUpdate = TRUE;
   settings->SupportDisplayControl = TRUE;
+  settings->RemoteFxCodec = TRUE;
+  settings->ColorDepth = 32;
   settings->RedirectClipboard = FALSE;
+  settings->SupportGraphicsPipeline = TRUE;
+
+  collections[0] = "disp";
+  freerdp_client_add_dynamic_channel (settings, count, collections);
+
+  build_options = g_ascii_strup (freerdp_get_build_config (), -1);
+  if (g_strrstr (build_options, "WITH_GFX_H264=ON") != NULL) {
+    settings->GfxH264 = TRUE;
+    settings->GfxAVC444 = TRUE;
+  } else {
+    settings->GfxH264 = FALSE;
+    settings->GfxAVC444 = FALSE;
+  }
+  g_free (build_options);
 
   settings->KeyboardLayout = freerdp_keyboard_init (0);
+
+  freerdp_register_addin_provider(freerdp_channels_load_static_addin_entry, 0);
 }
 
 static void
@@ -610,6 +817,9 @@ frdp_session_get_property (GObject    *object,
       case PROP_SCALING:
         g_value_set_boolean (value, self->priv->scaling);
         break;
+      case PROP_MONITOR_LAYOUT_SUPPORTED:
+        g_value_set_boolean (value, self->priv->monitor_layout_supported);
+        break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
         break;
@@ -647,6 +857,10 @@ frdp_session_set_property (GObject      *object,
       case PROP_SCALING:
         frdp_session_set_scaling (self, g_value_get_boolean (value));
         break;
+      case PROP_MONITOR_LAYOUT_SUPPORTED:
+        self->priv->monitor_layout_supported = g_value_get_boolean (value);
+        g_object_notify (G_OBJECT (self), "monitor-layout-supported");
+        break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
         break;
@@ -720,6 +934,14 @@ frdp_session_class_init (FrdpSessionClass *klass)
                                                          TRUE,
                                                          G_PARAM_READWRITE));
 
+  g_object_class_install_property (gobject_class,
+                                   PROP_MONITOR_LAYOUT_SUPPORTED,
+                                   g_param_spec_boolean ("monitor-layout-supported",
+                                                         "monitor-layout-supported",
+                                                         "monitor-layout-supported",
+                                                         FALSE,
+                                                         G_PARAM_READWRITE));
+
   signals[RDP_ERROR] = g_signal_new ("rdp-error",
                                      FRDP_TYPE_SESSION,
                                      G_SIGNAL_RUN_FIRST,
@@ -854,8 +1076,8 @@ frdp_session_mouse_event (FrdpSession          *self,
   input = priv->freerdp_session->input;
 
   if (priv->scaling) {
-    x = (x - priv->offset_x) / priv->scale_x;
-    y = (y - priv->offset_y) / priv->scale_y;
+    x = (x - priv->offset_x) / priv->scale;
+    y = (y - priv->offset_y) / priv->scale;
   }
 
   x = x < 0.0 ? 0.0 : x;
